from pytest_cases import fixture_ref, parametrize_plus

from carnival import cmd


@parametrize_plus('host_context', [
    fixture_ref('local_host_connection_context'),
    fixture_ref('ubuntu_ssh_host_connection'),
    fixture_ref('centos_ssh_host_connection'),
])
def test_is_dir_exists(suspend_capture, host_context):
    with suspend_capture:
        assert cmd.fs.is_dir_exists("/etc")
        assert cmd.fs.is_dir_exists("/bin")


@parametrize_plus('host_context', [
    fixture_ref('ubuntu_ssh_host_connection'),
    fixture_ref('centos_ssh_host_connection'),
])
def test_mkdirs(suspend_capture, host_context):
    with suspend_capture:
        assert cmd.fs.is_dir_exists("/tmp/.carnivaltestdir1") is False
        assert cmd.fs.is_dir_exists("/tmp/.carnivaltestdir2") is False

        cmd.fs.mkdirs("/tmp/.carnivaltestdir1", "/tmp/.carnivaltestdir2")

        assert cmd.fs.is_dir_exists("/tmp/.carnivaltestdir1") is True
        assert cmd.fs.is_dir_exists("/tmp/.carnivaltestdir2") is True

        cmd.cli.run(f"rm -rf /tmp/.carnivaltestdir1 /tmp/.carnivaltestdir2")



from pytest_cases import fixture_ref, parametrize_plus

from carnival import cmd


@parametrize_plus('host_context', [
    fixture_ref('ubuntu_ssh_host_connection'),
])
def test_install_ce_ubuntu(suspend_capture, host_context):
    with suspend_capture:
        assert cmd.fs.is_file_exists("/usr/bin/docker") is False
        cmd.docker.install_ce_ubuntu()
        assert cmd.fs.is_file_exists("/usr/bin/docker") is True
        cmd.apt.remove("docker-ce")


@parametrize_plus('host_context', [
    fixture_ref('ubuntu_ssh_host_connection'),
    fixture_ref('centos_ssh_host_connection'),
])
def test_install_compose(suspend_capture, host_context):
    with suspend_capture:
        assert cmd.fs.is_file_exists("/usr/local/bin/docker-compose") is False
        cmd.docker.install_compose()
        assert cmd.fs.is_file_exists("/usr/local/bin/docker-compose") is True

        cmd.cli.run("rm /usr/local/bin/docker-compose")



from invoke import Context

from carnival import global_context, Host


def test_global_context():
    h = Host("local")
    global_context.set_context(h)
    assert global_context.host == h
    assert global_context.conn is not None
    assert isinstance(global_context.conn, Context)



from carnival import utils, Host, global_context


def test_log(capsys):
    global_context.set_context(Host("local"))
    utils.log("Hellotest")

    captured = capsys.readouterr()
    assert captured.out == "💃💃💃 local> Hellotest\n"



from carnival import Task


class Help(Task):
    def run(self):
        from carnival.cli import get_tasks
        tasklist = list(get_tasks().keys())
        tasklist.sort()

        for task in tasklist:
            print(f" * {task}")



from carnival import cmd
from carnival.utils import log


def install_ce_ubuntu(version=None) -> bool:
    """
    Returns true if installed, false if was already installed
    """
    # https://docs.docker.com/v17.09/engine/installation/linux/docker-ce/ubuntu/
    from carnival.cmd import apt
    pkgname = "docker-ce"
    if apt.is_pkg_installed(pkgname, version):
        log(f"{pkgname} already installed")
        return False

    log(f"Installing {pkgname}...")

    cmd.cli.run("sudo apt-get remove docker docker-engine docker.io")
    cmd.cli.run("sudo apt-get update")
    cmd.cli.run("sudo apt-get install -y apt-transport-https ca-certificates curl software-properties-common")
    cmd.cli.run("curl -fsSL https://download.docker.com/linux/ubuntu/gpg | sudo apt-key add -", pty=True)
    cmd.cli.run('sudo add-apt-repository "deb [arch=amd64] https://download.docker.com/linux/ubuntu $(lsb_release -cs) stable"')

    apt.force_install(pkgname=pkgname, version=version, update=True, hide=True)
    return True


def install_compose(version="1.25.1", dest="/usr/local/bin/docker-compose"):
    log(f"Installing compose...")
    download_link = f"https://github.com/docker/compose/releases/download/{version}/docker-compose-`uname -s`-`uname -m`"
    cmd.cli.run(f"sudo curl -sL {download_link} -o {dest}")
    cmd.cli.run(f"sudo chmod a+x {dest}")



from pytest_cases import fixture_ref, parametrize_plus

from carnival import cmd


@parametrize_plus('host_context', [
    fixture_ref('local_host_connection_context'),
    fixture_ref('ubuntu_ssh_host_connection'),
    fixture_ref('centos_ssh_host_connection'),
])
def test_run(suspend_capture, host_context):
    with suspend_capture:
        result = cmd.cli.run("ls -1 /", hide=True)
        assert result.ok is True

        # Check response looks like root fs
        root_files = result.stdout.split("\n")
        assert 'bin' in root_files
        assert 'etc' in root_files
        assert 'usr' in root_files
        assert 'tmp' in root_files
        assert 'sbin' in root_files


@parametrize_plus('host_context', [
    fixture_ref('local_host_connection_context'),
    fixture_ref('ubuntu_ssh_host_connection'),
    fixture_ref('centos_ssh_host_connection'),
])
def test_pty(suspend_capture, host_context):
    with suspend_capture:
        result = cmd.cli.pty("ls -1 / | grep bin", hide=True)
        assert result.ok is True

        # Check response looks like root fs, filtered 'bin'
        root_files = result.stdout.split("\n")
        assert 'bin\r' in root_files
        assert 'sbin\r' in root_files



import os

from jinja2 import Environment, FileSystemLoader

from carnival import global_context
from carnival.secrets_manager import secrets_storage

j2_env = Environment(
    loader=FileSystemLoader(os.getcwd())
)


def render(template_path: str, **context) -> str:
    template = j2_env.get_template(template_path)
    return template.render(
        conn=global_context.conn,
        connected_host=global_context.host,
        host_context=global_context,
        secrets=secrets_storage,
        **context,
    )



from carnival import cli


def test_load_tasks_file():
    assert cli.load_tasks_file("not_exists_file123_") == {}

    tasks = cli.load_tasks_file("tests/fixtures/carnival_file_test.py")
    assert len(tasks) == 1



from typing import Union

from fabric import Connection  # type:ignore
from invoke import Context  # type:ignore

from carnival.host import Host


# noinspection PyTypeChecker
conn: Union[Connection, Context] = None
# noinspection PyTypeChecker
host: Host = None  # type:ignore


def set_context(h: Host):
    global conn
    global host

    assert host is None, f"Cannot set context, while other context active: {host}"
    assert conn is None, f"Cannot set context, while other context active: {conn}"

    conn = h.connect()
    host = h


def flush_context():
    global conn
    global host
    conn = None
    host = None



import os
import sys
import abc
from typing import Iterable, Type, Dict, Set

import click

from carnival.task import Task


def task_subclasses(cls) -> Set[Type]:
    # Get subclasses of task, which not abstract

    subclasses = set()
    for sc in cls.__subclasses__():
        # Skip if last MRO base is ABC
        if abc.ABC != sc.__mro__[1]:
            subclasses.add(sc)

        subclasses.update(task_subclasses(sc))

    return subclasses


def get_task_full_name(carnival_tasks_module: str, task_class: Type[Task]) -> str:
    task_name = task_class.get_name()

    task_mod = task_class.__module__

    # Internal tasks always in root
    if task_mod.startswith("carnival."):
        return task_name

    task_full_name = f"{task_mod}.{task_name}"

    if task_mod.startswith(carnival_tasks_module):
        task_full_name = task_full_name[len(carnival_tasks_module) + 1:]

    return task_full_name


def load_tasks_file(carnival_tasks_module: str) -> Dict[str, Type[Task]]:
    try:
        __import__(carnival_tasks_module)
    except (ModuleNotFoundError, FileNotFoundError) as ex:
        print(f"Cannot import {carnival_tasks_module}: {ex}", file=sys.stderr)

    tasks: Dict[str, Type[Task]] = {}

    for task_class in task_subclasses(Task):
        tasks[get_task_full_name(carnival_tasks_module, task_class)] = task_class

    return tasks


def get_tasks() -> Dict[str, Type[Task]]:
    sys.path.insert(0, os.getcwd())
    from carnival import internal_tasks  # noqa
    carnival_tasks_module = os.getenv("CARNIVAL_TASKS_MODULE", "carnival_tasks")
    return load_tasks_file(carnival_tasks_module)


def main():
    task_types = get_tasks()

    @click.command()
    @click.option('-d', '--dry_run', is_flag=True, default=False, help="Simulate run")
    @click.argument('tasks', required=True, type=click.Choice(task_types.keys()), nargs=-1)
    def cli(dry_run: bool, tasks: Iterable[str]):
        for task in tasks:
            task_types[task](dry_run=dry_run).run()
    cli()



import inspect
from typing import List, Dict, Any

from carnival.host import Host

from carnival.secrets_manager import secrets_storage


def _get_arg_names(fn) -> List[str]:
    arg_names: List[str] = []
    spec = inspect.getfullargspec(fn)
    arg_names += spec.args
    arg_names += spec.kwonlyargs
    return arg_names


def _build_kwargs(fn, context: Dict[str, Any]):
    arg_names: List[str] = _get_arg_names(fn)
    kwargs = {}

    for context_name, context_val in context.items():
        if context_name in arg_names:
            kwargs[context_name] = context_val
    return kwargs


class Step:
    def __init__(self, **context):
        self.context = context

    def _build_context(self, host: Host) -> Dict[str, Any]:
        run_context = {'secrets': secrets_storage, 'host': host}
        if host.context:
            run_context.update(host.context)
        if self.context:
            run_context.update(self.context)
        return run_context

    def run_with_context(self, host: Host):
        context = self._build_context(host)
        kwargs = _build_kwargs(self.run, context)
        return self.run(**kwargs)

    def run(self, **kwargs):
        raise NotImplementedError



from carnival.cmd import (
    cli,
    system,
    systemd,
    apt,
    docker,
    transfer,
    fs,
)

__all__ = [
    'cli',
    'system',
    'systemd',
    'apt',
    'docker',
    'transfer',
    'fs',
]
